/*****************************************************************
 * BotInterface.cpp
 * 
 * This file contains the code necessary to interface with the
 * FlexBot module in Half-Life. If you are writing a bot
 * behavior, there is no need to modify this file. All behavior-
 * specific code should be placed in the "npc.cpp" and "npc.h"
 * files.
 *
 * This implementation of the behavior interface is a simple yet
 * robust way to make sure that your FlexBot behavior can control
 * multiple bots simultanously without stomping on other bots'
 * state information. This is achieved by creating a new npc
 * object for each active bot.
 *****************************************************************/

#include <windows.h>
#include "stdio.h"
#include "sys/timeb.h"
#include "BehaviorInterface.h"
#include "grl-agent.h"
#include "npc.h"


/*****************************************************
 * COLLECTION OF BOTS USING THIS BEHAVIOR		     *
 *													 *
 * There is a grl_agent object for each instance of  *
 * a bot using the behavior.						 *
 *****************************************************/
grl_agent * bots[MAX_BOTS];
bool botArrayInitialized = false;

/*****************************************************
 * POINTERS TO SENSOR FUNCTIONS						 *
 *													 *
 * These functions are exported by the FlexBot MOD   *
 * and are used by wrapper functions to make sensor  *
 * calls.											 *
 *****************************************************/
PFN_VISION_DIST pfnGetVisionDistance;
PFN_VISION_DIST_WITH_PITCH pfnGetVisionDistanceWithPitch;
PFN_VISION_SWEEP_DIST pfnGetVisionSweepDistance;
PFN_PREDEF_VISION_DIST pfnGetLeftDistance;
PFN_PREDEF_VISION_DIST pfnGetCenterDistance;
PFN_PREDEF_VISION_DIST pfnGetRightDistance;
PFN_YAW_TO_PLAYER pfnYawToPlayer;
PFN_DIST_TO_PLAYER pfnDistanceToPlayer;
PFN_IS_PLAYER_VISIBLE pfnIsPlayerVisible;
PFN_PITCH_TO_PLAYER pfnPitchToPlayer;
PFN_HEALTH pfnGetHealth;
PFN_ROTATION pfnGetYaw;
PFN_ROTATION pfnGetPitch;
PFN_PLAYER_INDEX_RETURN pfnNearestVisiblePlayer;
PFN_IS_MOVING pfnIsMoving;
PFN_DEBUG_SIGNAL pfnDebugSignal;
PFN_DIST_TO_WAYPOINT pfnDistanceToWaypoint;
PFN_YAW_TO_WAYPOINT pfnYawToWaypoint;
PFN_NEAREST_WAYPOINT pfnNearestWaypoint;
PFN_DIST_TO_INFOPOINT pfnDistanceToInfopoint;
PFN_YAW_TO_INFOPOINT pfnYawToInfopoint;
PFN_NEAREST_INFOPOINT pfnNearestInfopoint;
PFN_NEXT_IP_TO_EXPIRE pfnNextIPToExpire;
PFN_AMMO_IN_CLIP pfnAmmoInClip;
PFN_NEAREST_VISIBLE_TEAMMATE pfnNearestVisibleTeammate;
PFN_YAW_TO_ENEMY_SOUND pfnYawToEnemySound;
PFN_DISTANCE_TO_ENEMY_SOUND pfnDistanceToEnemySound;
PFN_NEAREST_ITEM pfnNearestItem;
PFN_YAW_TO_ITEM pfnYawToItem;
PFN_DISTANCE_TO_ITEM pfnDistanceToItem;
PFN_PITCH_TO_ITEM pfnPitchToItem;
PFN_CAN_GET_ITEM pfnCanGetItem;
PFN_GET_CURRENT_WEAPON pfnGetCurrentWeapon;
PFN_GET_X pfnGetX; 
PFN_GET_Y pfnGetY;
PFN_GET_Z pfnGetZ;
PFN_IS_FRIEND pfnIsFriend;
PFN_PITCH_TO_INFOPOINT pfnPitchToInfopoint;
PFN_PITCH_TO_WAYPOINT pfnPitchToWaypoint;
PFN_GET_POSSESSES_GUN pfnGetPossessesGun;
PFN_PRIMARY_AMMO_IN_STOCK pfnPrimaryAmmoInStock;
PFN_SECONDARY_AMMO_IN_STOCK pfnSecondaryAmmoInStock;
PFN_ON_A_LADDER pfnOnALadder;
PFN_ADD_WAYPOINT pfnAddWaypoint;
PFN_ADD_INFOPOINT pfnAddInfopoint;


/*****************************************************
 * BOT MANIPULATION FUNCTIONS					     *
 *													 *
 * These functions are used to achieve tasks such as *
 * adding bots, removing bots, and bot thinking.	 *
 *****************************************************/

// Add a new instance of a bot using this behavior
EXPORT_BOT int CreateBotBehavior()
{
	// Initialize the bot array if this is the first bot
	if (!botArrayInitialized)
	{
		for (int i = 0; i < MAX_BOTS; i++)
			bots[i] = NULL;
		botArrayInitialized = true;
	}

	// Find the first open slot in the bot array
	for (int slot = 0; slot < MAX_BOTS; slot++)
		if (bots[slot] == NULL)
			break;

	// Return -1 if a slot could not be found
	if (bots[slot] != NULL)
		return -1;

	// Create a new npc object and store it
	grl_agent * newBot = new npc();
	bots[slot] = newBot;
	
	return slot;
}

// Checks if a bot instance can be safely accessed
bool isValidBot(int instance)
{
	// Is the bot in a valid range?
	if (instance < 0 && instance >= MAX_BOTS)
		return false;

	// Is there an instance of this bot?
	else if (bots[instance] == NULL)
		return false;

	return true;
}

// Remove an instance of a bot using this behavior
EXPORT_BOT void RemoveBotBehavior(int instance)
{
	if (isValidBot(instance))
	{
		delete bots[instance];
		bots[instance] = NULL;
	}
}

// Call the think function for the proper instance
EXPORT_BOT void Think(int instance, int clock)
{
	if (isValidBot(instance))
		// Call the think function
		bots[instance]->update(clock);
}

/*****************************************************
 * ACTUATOR ACCESSOR FUNCTIONS (EXPORTED)			 *
 *													 *
 * These functions will be called by the FlexBot MOD *
 * to get the actuator values to be applied for the  *
 * current think cycle.								 *
 *****************************************************/
EXPORT_BOT int GetRotate(int instance)
{
	if (!isValidBot(instance))
		return 0;

	return bots[instance]->rotate;
}

EXPORT_BOT int GetTranslate(int instance)
{
	if (!isValidBot(instance))
		return 0;

	return bots[instance]->translate;
}

EXPORT_BOT int GetStrafe(int instance)
{
	if (!isValidBot(instance))
		return 0;

	return bots[instance]->strafe;
}

EXPORT_BOT int GetPitch(int instance)
{
	if (!isValidBot(instance))
		return 0;

	return bots[instance]->pitch;
}

EXPORT_BOT bool GetShoot(int instance)
{
	if (!isValidBot(instance))
		return false;

	return bots[instance]->shoot;
}

EXPORT_BOT bool GetDuck(int instance)
{
	if (!isValidBot(instance))
		return false;

	return bots[instance]->duck;
}

EXPORT_BOT bool GetJump(int instance)
{
	if (!isValidBot(instance))
		return false;

	return bots[instance]->jump;
}

EXPORT_BOT bool GetSecondShoot(int instance)
{
	if (!isValidBot(instance))
		return false;

	return bots[instance]->secondary;
}

EXPORT_BOT int GetWeapon(int instance)
{
	if (!isValidBot(instance))
		return false;

	return bots[instance]->weapon;
}

EXPORT_BOT bool GetReload(int instance)
{
	if (!isValidBot(instance))
		return false;

	return bots[instance]->reload;
}

EXPORT_BOT bool GetUse(int instance)
{
	if (!isValidBot(instance))
		return false;

	return bots[instance]->use;
}

/*****************************************************
 * BEHAVIOR DLL INITIALIZATION						 *
 *													 *
 * These functions are called when the behavior DLL	 *
 * is first loaded. They ensure that the FlexBot DLL *
 * provides the required sensors and debug functions.*
 *****************************************************/

// Load and verify the required sensors
bool InitSensors()
{
	// Load the DLL -- note that the path must be accurate...
	// the path is relative to the active (running) Half-Life directory
	HINSTANCE hInstDll = LoadLibrary("FlexBot/dlls/FlexBot.dll");
	if (hInstDll)
	{
		// Load the functions and return failure if we can't get one (incompatible w/ main module)
		pfnGetVisionDistance = (PFN_VISION_DIST) GetProcAddress(hInstDll, "BOTEXP_GetVisionDistance");
		if (!pfnGetVisionDistance)
			return TRUE;
		pfnGetVisionDistanceWithPitch = (PFN_VISION_DIST_WITH_PITCH) GetProcAddress(hInstDll, "BOTEXP_GetVisionDistanceWithPitch");
		if (!pfnGetVisionDistanceWithPitch)
			return TRUE;
		pfnGetVisionSweepDistance = (PFN_VISION_SWEEP_DIST) GetProcAddress(hInstDll, "BOTEXP_GetVisionSweepDistance");
		if (!pfnGetVisionSweepDistance)
			return TRUE;
		pfnGetLeftDistance = (PFN_PREDEF_VISION_DIST) GetProcAddress(hInstDll, "BOTEXP_GetLeftDistance");
		if (!pfnGetLeftDistance)
			return TRUE;
		pfnGetRightDistance = (PFN_PREDEF_VISION_DIST) GetProcAddress(hInstDll, "BOTEXP_GetRightDistance");
		if (!pfnGetRightDistance)
			return TRUE;
		pfnGetCenterDistance = (PFN_PREDEF_VISION_DIST) GetProcAddress(hInstDll, "BOTEXP_GetCenterDistance");
		if (!pfnGetCenterDistance)
			return TRUE;
		pfnYawToPlayer = (PFN_YAW_TO_PLAYER) GetProcAddress(hInstDll, "BOTEXP_YawToPlayer");
		if (!pfnYawToPlayer)
			return TRUE;
		pfnDistanceToPlayer = (PFN_DIST_TO_PLAYER) GetProcAddress(hInstDll, "BOTEXP_DistanceToPlayer");
		if (!pfnDistanceToPlayer)
			return TRUE;
		pfnIsPlayerVisible = (PFN_IS_PLAYER_VISIBLE) GetProcAddress(hInstDll, "BOTEXP_IsPlayerVisible");
		if (!pfnIsPlayerVisible)
			return TRUE;
		pfnPitchToPlayer = (PFN_PITCH_TO_PLAYER) GetProcAddress(hInstDll, "BOTEXP_PitchToPlayer");
		if (!pfnPitchToPlayer)
			return TRUE;
		pfnGetHealth = (PFN_HEALTH) GetProcAddress(hInstDll, "BOTEXP_GetHealth");
		if (!pfnGetHealth)
			return TRUE;
		pfnGetYaw = (PFN_ROTATION) GetProcAddress(hInstDll, "BOTEXP_GetYaw");
		if (!pfnGetYaw)
			return TRUE;
		pfnGetPitch = (PFN_ROTATION) GetProcAddress(hInstDll, "BOTEXP_GetPitch");
		if (!pfnGetPitch)
			return TRUE;
		pfnNearestVisiblePlayer = (PFN_PLAYER_INDEX_RETURN) GetProcAddress(hInstDll, "BOTEXP_NearestVisEnemyNum");
		if (!pfnNearestVisiblePlayer)
			return TRUE;
		pfnIsMoving = (PFN_IS_MOVING) GetProcAddress(hInstDll, "BOTEXP_IsMoving");
		if(!pfnIsMoving)
			return TRUE;
		pfnDebugSignal = (PFN_DEBUG_SIGNAL) GetProcAddress(hInstDll, "BOTEXP_DebugSignal");
		if (!pfnDebugSignal)
			return TRUE;
		pfnDistanceToWaypoint = (PFN_DIST_TO_WAYPOINT) GetProcAddress(hInstDll, "BOTEXP_DistanceToWaypoint");
		if (!pfnDistanceToWaypoint)
			return TRUE;
		pfnYawToWaypoint = (PFN_YAW_TO_WAYPOINT) GetProcAddress(hInstDll, "BOTEXP_YawToWaypoint");
		if (!pfnYawToWaypoint)
			return TRUE;
		pfnNearestWaypoint = (PFN_NEAREST_WAYPOINT) GetProcAddress(hInstDll, "BOTEXP_NearestWaypoint");
		if (!pfnNearestWaypoint)
			return TRUE;
		pfnDistanceToInfopoint = (PFN_DIST_TO_INFOPOINT) GetProcAddress(hInstDll, "BOTEXP_DistanceToInfopoint");
		if (!pfnDistanceToInfopoint)
			return TRUE;
		pfnYawToInfopoint = (PFN_YAW_TO_INFOPOINT) GetProcAddress(hInstDll, "BOTEXP_YawToInfopoint");
		if (!pfnYawToInfopoint)
			return TRUE;
		pfnNearestInfopoint = (PFN_NEAREST_INFOPOINT) GetProcAddress(hInstDll, "BOTEXP_NearestInfopoint");
		if (!pfnNearestInfopoint)
			return TRUE;
		pfnNextIPToExpire = (PFN_NEXT_IP_TO_EXPIRE) GetProcAddress(hInstDll, "BOTEXP_NextIPToExpire");
		if (!pfnNextIPToExpire)
			return TRUE;
		pfnAmmoInClip = (PFN_AMMO_IN_CLIP) GetProcAddress(hInstDll, "BOTEXP_AmmoInClip");
		if (!pfnAmmoInClip)
			return TRUE;
		pfnNearestVisibleTeammate = (PFN_NEAREST_VISIBLE_TEAMMATE) GetProcAddress(hInstDll, "BOTEXP_NearestVisTeammateNum");
		if (!pfnNearestVisibleTeammate)
			return TRUE;
		pfnYawToEnemySound = (PFN_YAW_TO_ENEMY_SOUND) GetProcAddress(hInstDll, "BOTEXP_YawToEnemySound");
		if (!pfnYawToEnemySound)
			return TRUE;
		pfnDistanceToEnemySound = (PFN_DISTANCE_TO_ENEMY_SOUND) GetProcAddress(hInstDll, "BOTEXP_DistanceToEnemySound");
		if (!pfnDistanceToEnemySound)
			return TRUE;
		pfnNearestItem = (PFN_NEAREST_ITEM) GetProcAddress(hInstDll, "BOTEXP_NearestItem");
		if (!pfnNearestItem)
			return TRUE;
		pfnYawToItem = (PFN_YAW_TO_ITEM) GetProcAddress(hInstDll, "BOTEXP_YawToItem");
		if (!pfnNearestItem)
			return TRUE;
		pfnDistanceToItem = (PFN_DISTANCE_TO_ITEM) GetProcAddress(hInstDll, "BOTEXP_DistanceToItem");
		if (!pfnNearestItem)
			return TRUE;
		pfnPitchToItem = (PFN_PITCH_TO_ITEM) GetProcAddress(hInstDll, "BOTEXP_PitchToItem");
		if (!pfnNearestItem)
			return TRUE;
		pfnCanGetItem = (PFN_CAN_GET_ITEM) GetProcAddress(hInstDll, "BOTEXP_CanGetItem");
		if (!pfnCanGetItem)
			return TRUE;
		pfnGetCurrentWeapon = (PFN_GET_CURRENT_WEAPON) GetProcAddress(hInstDll, "BOTEXP_GetCurrentWeapon");
		if (!pfnGetCurrentWeapon)
			return TRUE;
		pfnGetX = (PFN_GET_X) GetProcAddress(hInstDll, "BOTEXP_GetX");
		if (!pfnGetX)
			return TRUE;
		pfnGetY= (PFN_GET_Y) GetProcAddress(hInstDll, "BOTEXP_GetY");
		if (!pfnGetY)
			return TRUE;
		pfnGetZ= (PFN_GET_Z) GetProcAddress(hInstDll, "BOTEXP_GetZ");
		if (!pfnGetZ)
			return TRUE;
		pfnIsFriend = (PFN_IS_FRIEND) GetProcAddress(hInstDll, "BOTEXP_IsFriend");
		if (!pfnIsFriend)
			return TRUE;
		pfnPitchToInfopoint = (PFN_PITCH_TO_INFOPOINT) GetProcAddress(hInstDll, "BOTEXP_PitchToInfopoint");
		if (!pfnPitchToInfopoint)
			return TRUE;
		pfnPitchToWaypoint = (PFN_PITCH_TO_WAYPOINT) GetProcAddress(hInstDll, "BOTEXP_PitchToWaypoint");
		if (!pfnPitchToWaypoint)
			return TRUE;
		pfnPrimaryAmmoInStock = (PFN_PRIMARY_AMMO_IN_STOCK) GetProcAddress(hInstDll, "BOTEXP_PrimaryAmmoInStock");
		if (!pfnPrimaryAmmoInStock )
			return TRUE;
		pfnSecondaryAmmoInStock = (PFN_SECONDARY_AMMO_IN_STOCK) GetProcAddress(hInstDll, "BOTEXP_SecondaryAmmoInStock");
		if (!pfnSecondaryAmmoInStock )
			return TRUE;
		pfnGetPossessesGun = (PFN_GET_POSSESSES_GUN) GetProcAddress(hInstDll, "BOTEXP_GetPossessesGun");
		if (!pfnGetPossessesGun)
			return TRUE;
		pfnOnALadder = (PFN_ON_A_LADDER) GetProcAddress(hInstDll, "BOTEXP_OnLadder");
		if (!pfnOnALadder)
			return TRUE;
		pfnAddWaypoint = (PFN_ADD_WAYPOINT) GetProcAddress(hInstDll, "BOTEXP_AddWaypoint");
		if (!pfnAddWaypoint)
			return TRUE;
		pfnAddInfopoint = (PFN_ADD_INFOPOINT) GetProcAddress(hInstDll, "BOTEXP_AddInfopoint");
		if (!pfnAddInfopoint)
			return TRUE;
	}
	else
	{
		MessageBox(NULL, "ERROR: Could not load the MOD DLL from the behavior DLL. Check the path in BehaviorInterface.cpp", "DLL Trouble", MB_OK);
		return TRUE;
	}
	return FALSE;
}

int WINAPI DllMain(HINSTANCE hInstance, DWORD fdwReason, PVOID pvReserved)
{
	if (InitSensors())
		return 0;

	return 1;
}

/*****************************************************
 * DEBUGGING (TRACE) FUNCTIONS						 *
 *													 *
 * These functions are called to pass debug info to  *
 * the FlexBot MOD. There a variety of things that	 *
 * the MOD may do with this debug info, including	 *
 * displaying the name/value pairs in the BotWatch	 *
 * debugging application.							 *
 *****************************************************/


void update_trace(const char * signalName, int value)
{	// integer traces
	char debugStr[256];
	sprintf(debugStr, "%s %d", signalName, value);
	(pfnDebugSignal)(debugStr);
}

void update_trace(const char * signalName, float value)
{	// float traces
	char debugStr[256];
	sprintf(debugStr, "%s %f", signalName, value);
	(pfnDebugSignal)(debugStr);
}

void update_trace(const char * signalName, bool value)
{	// boolean traces
	char debugStr[256];
	if (value)
		sprintf(debugStr, "%s TRUE", signalName);
	else
		sprintf(debugStr, "%s FALSE", signalName);
	(pfnDebugSignal)(debugStr);
}


/*****************************************************
 * WRAPPER FUNCTIONS FOR SENSORS					 *
 *													 *
 * These functions provide a convenient interface    *
 * for calling sensor functions without having to	 *
 * deal with function pointers, which can be		 *
 * dangerous because there is no type-checking or	 *
 * argument-count checking for calls to raw function *
 * pointers. This prevents incorrect usage, which	 *
 * often leads to crashes.							 *
 *****************************************************/
int get_vision_distance(float angle, int range)
{
	return (pfnGetVisionDistance)(angle, range);
}
int get_vision_distance_with_pitch(float angle, float pitch, int range)
{
	return (pfnGetVisionDistanceWithPitch)(angle, pitch, range);
}
int get_vision_sweep_distance(float angle, int range, int sweep, int interval)
{
	return (pfnGetVisionSweepDistance)(angle, range, sweep, interval);
}
int get_left_distance(int range)
{
	return (pfnGetLeftDistance)(range);
}
int get_right_distance(int range)
{
	return (pfnGetRightDistance)(range);
}
int get_center_distance(int range)
{
	return (pfnGetCenterDistance)(range);
}
float yaw_to_player(int playerNum)
{
	return (pfnYawToPlayer)(playerNum);
}
float pitch_to_player(int playerNum)
{
	return (pfnPitchToPlayer)(playerNum);
}
int distance_to_player(int playerNum)
{
	return (pfnDistanceToPlayer)(playerNum);
}
bool is_player_visible(int playerNum)
{
	return (pfnIsPlayerVisible)(playerNum);
}
int get_health()
{
	return (pfnGetHealth)();
}
float get_absolute_yaw()
{
	return (pfnGetYaw)();
}
float get_absolute_pitch()
{
	return (pfnGetPitch)();
}
int nearest_visible_enemy()
{
	return (pfnNearestVisiblePlayer)();
}
bool is_moving()
{
	return (pfnIsMoving)();
}
float yaw_to_waypoint(int waypointNum)
{
	return (pfnYawToWaypoint)(waypointNum);
}
int distance_to_waypoint(int waypointNum)
{
	return (pfnDistanceToWaypoint)(waypointNum);
}
int nearest_waypoint()
{
	return (pfnNearestWaypoint)();
}
float yaw_to_infopoint(int infopointNum, int type)
{
	return (pfnYawToInfopoint)(infopointNum, type);
}
int distance_to_infopoint(int infopointNum, int type)
{
	return (pfnDistanceToInfopoint)(infopointNum, type);
}
int nearest_infopoint(int type)
{
	return (pfnNearestInfopoint)(type);
}
int next_ip_to_expire(int type)
{
	return (pfnNextIPToExpire)(type);
}
int ammo_in_clip()
{
	return (pfnAmmoInClip)();
}
int nearest_visible_teammate()
{
	return (pfnNearestVisibleTeammate)();
}
float yaw_to_enemy_sound()
{
	return(pfnYawToEnemySound)();
}
int distance_to_enemy_sound()
{
	return(pfnDistanceToEnemySound)();
}
int nearest_item(int itemType)
{
	return(pfnNearestItem)(itemType);
}
float yaw_to_item(int itemType, int index)
{
	return (pfnYawToItem)(itemType, index);
}
int distance_to_item(int itemType, int index)
{
	return (pfnDistanceToItem)(itemType, index);
}
float pitch_to_item(int itemType, int index)
{
	return (pfnPitchToItem)(itemType, index);
}
bool can_get_item(int itemType, int index)
{
	return (pfnCanGetItem)(itemType, index);
}
int get_current_weapon()
{
	return (pfnGetCurrentWeapon)();
}
float pitch_to_infopoint(int pointType, int index)
{
	return (pfnPitchToInfopoint)(pointType, index);
}
float pitch_to_waypoint(int index)
{
	return (pfnPitchToWaypoint)(index);
}
bool is_friend(int index){
	return (pfnIsFriend)(index);
}
int get_x()
{
	return (pfnGetX)();
}
int get_y()
{
	return (pfnGetY)();
}
int get_z()
{
	return (pfnGetZ)();
}
bool get_possesses_gun(int gunID)
{
	return (pfnGetPossessesGun)(gunID);
}
int primary_ammo_in_stock(int gunID)
{
	return (pfnPrimaryAmmoInStock)(gunID);
}
int secondary_ammo_in_stock(int gunID)
{
	return (pfnSecondaryAmmoInStock)(gunID);
}
bool on_a_ladder()
{
	return (pfnOnALadder)();
}
void add_waypoint()
{
	(pfnAddWaypoint)();
}
void add_infopoint(int lifespan)
{
	(pfnAddInfopoint)(lifespan);
}